from binascii import b2a_base64, a2b_base64
from UserList import UserList
import simplejson as json

# maximum length of a single message
MAX_LENGTH=600

# the various types of messages
STRING_TYPE = 1     # just a string
MSGID_TYPE = 2      # a message id which needs to be translated

ALL_TYPES = (STRING_TYPE, MSGID_TYPE)

class Message(object):
    """A Message object can store a message and any additional 
    attributes as long as they can be encoded in JSON format.
    Per default it only takes the message but you can add additional
    data via keyword arguments like this::
    
        msg = Message(u'my message', typ = 1, format='restructuredText')
        
    """

    def __init__(self, msg=u'', **kws):
        """initialize the message. ``msg`` MUST be unicode."""
        self.msg = msg
        # store additional keywords in the class
        for a,v in kws.items():
            setattr(self,a,v)

    def __repr__(self):
        return "<Message '%s'>" %self.msg

    def encode(self):
        """encoding a message basically means returning the instance dict"""
        return self.__dict__

    def decode(self, d):
        """decode a dictionary into this object"""
        self.__dict__ = d

class MessageList(UserList):
    """a list of messages

    You can store more than one message as a statusmessage. You simply store them
    inside this list like object.
    """

    def encode(self):
        """return a representation of all the messages included. 
        This is created by using this list and encoding it via JSON and base64.

        If the list is empty we return None so that the middleware can eventually
        delete the cookie.
        """
        if len(self) == 0:
            return None
        l = []
        for msg in self:
            l.append(msg.encode())

        return b2a_base64(json.dumps(l)).rstrip()

def decode(s):
    """decode a list of messages

    This returns a ``MessageList`` object which consists of individual
    ``Message`` object.

    First is decodes base64 and then the json string inside creating
    individual ``Message`` objects out of it.

    """

    # first we get a list of dictionaries
    try:
        l = json.loads(a2b_base64(s))
    except ValueError:
        # sometimes cookies are set wrong (e.g. with IE), then we ignore it and set it to an empty list
        l = []

    # and now we convert those into Message objects
    msglist = MessageList()
    for d in l:
        m = Message()
        m.decode(d)
        msglist.append(m)
    return msglist




import time
from paste.request import get_cookies

from message import MessageList, decode


class StatusMessageMiddleware(object):
    """WSGI middleware for handling status messages in cookies

       Imagine you want to show some message like "blog has been added" 
       to the user after he hit "submit" on the addform. You most likely 
       want to do a redirect on a successful addition so that the user 
       cannot accidently hit "reload" and submit the data again.

       In order to do that this middleware offers you a possibility of
       storing one or more messages in a cookie. Therefor it provides
       a ``MessageList`` object in the WSGI environment under the key
       ``qc.statusmessage``. Check out the 
       ``qc.statusmessage.message.MessageList`` API for more information
       on how to use it but basically it's a simple list object and all
       methods usually available to lists apply.

       The contents of the ``MessageList`` object are serialized to a 
       cookie on egress and deserialized again on ingress. You can name
       the cookie in the constructor of this middleware.

    """

    def __init__(self, application, cookie_name="status"):
        """initialize the middleware with the application and a name
        of the cookie.
        """

        self.application = application
        self.cookie_name = cookie_name

    def __call__(self, environ, start_response):
        """On ingress lookup the cookie which is named as the
        ``self.cookie_name`` variable and deserialize the data
        into a ``MessageList`` object. This object is then put into
        the WSGI environment under ``qc.statusmessage``.

        On egress it will be done the other way round: The ``MessageList``
        object in the WSGI environment will be serialized to a cookie with
        it's name stored in ``self.cookie_name``.

        """

        cookies = get_cookies(environ)
        cookie = cookies.get(self.cookie_name, None)
        cookie_value = getattr(cookie,'value',None)
        if cookie_value is not None:
            msglist = decode(cookie_value)
        else:
            # create a new object from scratch
            msglist = MessageList()

        # place the message list in the environment
        environ['qc.statusmessage'] = msglist

        # create a replacement start_response function
        def replacement_start_response(status, headers, *args, **kwargs):
            """add the cookie to it on egress

            We still have the ``MessageList`` object available
            
            """
            cookie_value = msglist.encode()
            if cookie_value is not None:
                set_cookie='%s=%s; Path=/;' %(self.cookie_name, cookie_value)
            else:
                # delete the cookie
                set_cookie='%s=; Path=/; max-age=0; expires=%s' %(self.cookie_name, time.time()-60)
            headers.append(('Set-Cookie', set_cookie))
            return start_response(status, headers, *args, **kwargs)

        # now call the application with our custom start response wrapper
        return self.application(environ, replacement_start_response)

def make_statusmessage_middleware(app, global_conf, **local_conf):
    cookie_name = local_conf.get('cookie_name', None)
    return StatusMessageMiddleware(app, cookie_name)




import unittest
from qc.statusmessage.message import Message, MessageList, decode


class TestMessage(unittest.TestCase):

    def test_create_message(self):
        """test creating a message"""
        msg = Message(u"Dies ist ein Test")
        self.assertEqual(msg.msg, u"Dies ist ein Test")

    def test_create_message_with_attributes(self):
        msg = Message(u"Dies ist ein Test", a=3, b="test", c=u"test2", d=3.2)
        self.assertEqual(msg.msg, u"Dies ist ein Test")
        self.assertEqual(msg.a, 3)
        self.assertEqual(msg.b, "test")
        self.assertEqual(msg.c, u"test2")
        self.assertEqual(msg.d, 3.2)

    def test_encode(self):
        """test encoding a message"""
        msg = Message(u"Dies ist ein Test")

        res = msg.encode()
        self.assertEqual(res,{'msg': u'Dies ist ein Test'})

    def test_decode(self):
        """test decoding a message"""
        d={'msg': u'foobar'}
        msg = Message()
        msg.decode(d)

        self.assertEqual(msg.msg, u"foobar")

    def test_list_encode(self):
        """test basic list encoding functionality"""
        msg1 = Message(u"Message 1")
        msg2 = Message(u"Message 2", c=3)
        msglist = MessageList()
        msglist.append(msg1)
        msglist.append(msg2)

        res = msglist.encode()

        self.assertEqual(res,u"W3sibXNnIjogIk1lc3NhZ2UgMSJ9LCB7Im1zZyI6ICJNZXNzYWdlIDIiLCAiYyI6IDN9XQ==")


    def test_list_decode(self):
        """test list decoding"""
        from binascii import b2a_base64
        from simplejson import dumps
        data = b2a_base64(dumps([{'msg': u'Message 1'},
                {'msg': u'Message 2', 'c':3, 'test': u'new stuff'}]))
        msglist = decode(data)
        self.assertEqual(len(msglist),2)
        msg1, msg2 = msglist

        self.assertEqual(msg1.msg, u"Message 1")
        self.assertEqual(msg2.msg, u"Message 2")
        self.assertEqual(msg2.c, 3)
        self.assertEqual(msg2.test, u"new stuff")


class TestMiddleware(unittest.TestCase):

    def _makeOne(self, *arg, **kw):
        from qc.statusmessage.middleware import StatusMessageMiddleware
        return StatusMessageMiddleware(*arg, **kw)

    def test_write_message(self):
        app = DummyApplication(msg='Message 1')
        start_response = DummyStartResponse()
        smm = self._makeOne(app,'cookie')

        env = {}
        result = smm(env, start_response)

        cookie_header = start_response.headers[0][1]
        self.assertEqual(cookie_header, u'cookie=W3sibXNnIjogIk1lc3NhZ2UgMSJ9XQ==; Path=/;')

    def test_read_message(self):
        """test if a set message can be read again"""

        # make the next request with this cookie
        cvalue = '''cookie="W3sibXNnIjogIk1lc3NhZ2UgMSJ9XQ=="'''

        app = DummyApplication()
        start_response = DummyStartResponse()
        smm = self._makeOne(app,'cookie')

        env = {'HTTP_COOKIE': cvalue}
        result = smm(env, start_response)
        self.failUnless('qc.statusmessage' in env.keys())

        msglist = env['qc.statusmessage']
        self.assertEqual(len(msglist),1)
        self.assertEqual(msglist.pop().msg,u'Message 1')

    def test_add_message(self):
        """test if a message can be added"""

        # make the next request with this cookie
        cvalue = '''cookie="W3sibXNnIjogIk1lc3NhZ2UgMSJ9XQ=="'''

        app = DummyApplication(msg=u"another one")
        start_response = DummyStartResponse()
        smm = self._makeOne(app,'cookie')

        env = {'HTTP_COOKIE': cvalue}
        result = smm(env, start_response)
        self.failUnless('qc.statusmessage' in env.keys())

        msglist = env['qc.statusmessage']
        self.assertEqual(len(msglist),2)
        self.assertEqual(msglist.pop().msg,u'another one')
        self.assertEqual(msglist.pop().msg,u'Message 1')
        self.assertEqual(len(msglist),0)

    def test_if_cookie_is_deleted_on_empty_list(self):
        # first round: put a message into a cookie and we tell the application
        # to consume it (e.g. pop() it)
        cvalue = '''cookie="W3sibXNnIjogIk1lc3NhZ2UgMSJ9XQ=="'''
        app = DummyApplication(consume=True)
        start_response = DummyStartResponse()
        smm = self._makeOne(app,'cookie')

        env = {'HTTP_COOKIE': cvalue}

        # here the application is called which is told to delete the message
        result = smm(env, start_response)

        # here we get the new cookie value
        cvalue = start_response.headers[0][1]
        self.failUnless(cvalue.startswith('cookie=; Path=/; max-age=0; expires'))


class DummyApplication(object):

    # TODO: find out why headers is not [] when using it this way
    #def __init__(self, msg=None, headers = [], status='200 OK'):
    def __init__(self, msg=None, status='200 OK', headers=None, consume=False):
        self.msg = msg
        self.status = status
        if headers is None:
            headers = []
        self.headers = headers
        self.consume = consume

    def __call__(self, environ, start_response):
        ml = environ['qc.statusmessage']
        if self.msg is not None:
            ml.append(Message(self.msg))
        if self.consume:
            ml.pop()
        self.environ = environ
        self.start_response = start_response(self.status, self.headers)
        return ['hello world']

class DummyStartResponse(object):
    def __call__(self, status, headers, exc_info=None):
        self.status = status
        self.headers = headers
        self.exc_info = exc_info




from setuptools import setup, find_packages
import os

version = '1.0b2'

here = os.path.abspath(os.path.dirname(__file__))
try:
    README = open(os.path.join(here, 'README.txt')).read()
    CHANGES = open(os.path.join(here, 'HISTORY.txt')).read()
except IOError:
    README = CHANGES = ''


setup(name='qc.statusmessage',
      version=version,
      description="WSGI Middleware for displaying status messages",
      long_description=README + '\n\n' +  CHANGES,
      # Get more strings from http://www.python.org/pypi?%3Aaction=list_classifiers
      classifiers=[
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Internet :: WWW/HTTP :: WSGI",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Middleware", 
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Topic :: Internet :: WWW/HTTP",

        ],
      keywords='statusmessage quantumcore wsgi middleware web framework',
      author='Christian Scholz',
      author_email='cs@comlounge.net',
      url='http://quantumcore.net',
      license='BSD License, http://quantumcore.org/license.html',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['qc'],
      zip_safe=False,
      test_suite="qc.statusmessage.tests",
      include_package_data=True,
      install_requires=[
          'setuptools',
          'paste',
          'simplejson',
          # -*- Extra requirements: -*-
      ],
      entry_points="""\
      [paste.filter_app_factory]
      middleware = qc.statusmessage.middleware:make_statusmessage_middleware
      """,
      )
